/*
 * Copyright 1993, 1995, 2002 Paul Mattes.  All Rights Reserved.
 *   Permission to use, copy, modify, and distribute this software and its
 *   documentation for any purpose and without fee is hereby granted,
 *   provided that the above copyright notice appear in all copies and that
 *   both that copyright notice and this permission notice appear in
 *   supporting documentation.
 *
 * x026 -- A Keypunch Simluator
 */

#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Core.h>
#include <X11/Shell.h>
#include <X11/Xatom.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Dialog.h>
#include <X11/Xaw/Porthole.h>

#define SLOW	75
#define FAST	25

static char		*programname;
static Widget		toplevel;
static XtAppContext	appcontext;
static Display		*display;
static int		default_screen;
static int		root_window;
static int		card_window;
static int		depth;
static XFontStruct	*fgfontinfo;
static XFontStruct	*bgfontinfo;
static Atom		a_delete_me;
static int		line_number = 100;
static Boolean		drawing = True;

/* Internal actions. */
static void do_nothing(int);
static void do_data(int);
static void do_multipunch(int);
static void do_left(int);
static void do_right(int);
static void do_home(int);
static void do_pan_right(int);
static void do_pan_left(int);
static void do_pan_up(int);
static void do_slam(int);
static void do_newcard(int);

static void save(int ignored);

/* Application resources. */
typedef struct {
	Pixel	foreground;
	Pixel	background;
	Pixel	cabinet;
	Pixel	cardcolor;
	Pixel	errcolor;
	char	*fgfontname;
	char	*bgfontname;
	Boolean	autonumber;
	Boolean	typeahead;
} AppRes, *AppResptr;

static AppRes appres;

/* Command-line options. */
static XrmOptionDescRec options[]= {
	{ "-fgfont",	".fgfont",	XrmoptionSepArg,	NULL },
	{ "-bgfont",	".bgfont",	XrmoptionSepArg,	NULL },
	{ "-nonumber",	".autoNumber",  XrmoptionNoArg,		"False" },
	{ "-typeahead",	".typeahead",	XrmoptionNoArg,		"True" },
	{ "-mono",	".cabinet",	XrmoptionNoArg,		"black" },
};

/* Resource list. */
#define offset(field) XtOffset(AppResptr, field)
static XtResource resources[] = {
	{ XtNforeground, XtCForeground, XtRPixel, sizeof(Pixel),
	offset(foreground), XtRString, "XtDefaultForeground" },
	{ XtNbackground, XtCBackground, XtRPixel, sizeof(Pixel),
	offset(background), XtRString, "XtDefaultBackground" },
	{ "cabinet", "Cabinet", XtRPixel, sizeof(Pixel),
	offset(cabinet), XtRString, "grey" },
	{ "cardColor", "CardColor", XtRPixel, sizeof(Pixel),
	offset(cardcolor), XtRString, "ivory" },
	{ "errColor", "ErrColor", XtRPixel, sizeof(Pixel),
	offset(errcolor), XtRString, "firebrick" },
	{ "fgfont", "FgFont", XtRString, sizeof(String),
	offset(fgfontname), XtRString, 0 },
	{ "bgfont", "BgFont", XtRString, sizeof(String),
	offset(bgfontname), XtRString, 0 },
	{ "autoNumber", "AutoNumber", XtRBoolean, sizeof(Boolean),
	offset(autonumber), XtRString, "True" },
	{ "typeahead", "Typeahead", XtRBoolean, sizeof(Boolean),
	offset(typeahead), XtRString, "False" },
};
#undef offset

/* Fallback resources. */
static String fallbacks[] = {
	"*fgfont:	9x15",
	"*bgfont:	5x8",
	"*pos.font:	6x13bold",
	"*dialog*value*font: fixed",
	"*font:		variable",
	"*cabinet:	grey",
	"*cardColor:	ivory",
	NULL
};

/* Xt actions. */
static void data(Widget, XEvent *, String *, Cardinal *);
static void delete_window(Widget, XEvent *, String *, Cardinal *);
static void home(Widget, XEvent *, String *, Cardinal *);
static void left(Widget, XEvent *, String *, Cardinal *);
static void next(Widget, XEvent *, String *, Cardinal *);
static void redraw(Widget, XEvent *, String *, Cardinal *);
static void right(Widget, XEvent *, String *, Cardinal *);
static void tab(Widget, XEvent *, String *, Cardinal *);
static void insert_selection(Widget, XEvent *, String *, Cardinal *);
static void confirm(Widget, XEvent *, String *, Cardinal *);

/* Xt callbacks. */
static void discard(Widget, XtPointer, XtPointer);

/* Actions. */
XtActionsRec actions[] = {
	{ "Data",	data },
	{ "DeleteWindow", delete_window },
	{ "Home",	home },
	{ "Left",	left },
	{ "Next",	next },
	{ "Redraw",	redraw },
	{ "Right",	right },
	{ "Tab",	tab },
	{ "insert-selection", insert_selection },
	{ "confirm",	confirm }
};
int actioncount = XtNumber(actions);

#include "x026.bm"

/* Forward references. */
static void define_widgets(void);
static void new_card(void);
static void save_popup(void);


int
main(int argc, char *argv[])
{
	char	buf[256];
	XtTranslations table;
	Pixmap icon;

	/* Figure out who we are */
	programname = strrchr(argv[0], '/');
	if (programname)
		++programname;
	else
		programname = argv[0];

	/* Initialze Xt and fetch our resources. */
	toplevel = XtVaAppInitialize(
	    &appcontext,
	    "X026",
	    options, XtNumber(options),
	    &argc, argv,
	    fallbacks,
	    XtNinput, True,
	    XtNallowShellResize, False,
	    NULL);
	if (argc > 1) {
		(void) sprintf(buf, "Usage: %s [options]", programname);
		XtError(buf);
		exit(1);
	}
	XtGetApplicationResources(toplevel, &appres, resources,
	    XtNumber(resources), 0, 0);

	/* Set up some globals. */
	display = XtDisplay(toplevel);
	default_screen = DefaultScreen(display);
	root_window = RootWindow(display, default_screen);
	depth = DefaultDepthOfScreen(XtScreen(toplevel));
	a_delete_me = XInternAtom(display, "WM_DELETE_WINDOW", False);

	/* Set up actions. */
	XtAppAddActions(appcontext, actions, actioncount);

	/* Load fonts. */
	fgfontinfo = XLoadQueryFont(display, appres.fgfontname);
	if (fgfontinfo == NULL)
		XtError("Can't load fgFont");
	bgfontinfo = XLoadQueryFont(display, appres.bgfontname);
	if (bgfontinfo == NULL)
		XtError("Can't load bgFont");

	/* Define the widgets. */
	define_widgets();

	/* Set up a cute (?) icon. */
	icon = XCreateBitmapFromData(display, XtWindow(toplevel),
	    (char *)x026_bits, x026_width, x026_height);
	XtVaSetValues(toplevel, XtNiconPixmap, icon, XtNiconMask, icon, NULL);

	/* Allow us to die gracefully. */
	XSetWMProtocols(display, XtWindow(toplevel), &a_delete_me, 1);
	table = XtParseTranslationTable(
	    "<Message>WM_PROTOCOLS: DeleteWindow()");
	XtOverrideTranslations(toplevel, table);

#if defined(SOUND) /*[*/
	/* Set up clicks. */
	audio_init();
#endif /*]*/

	new_card();

	/* Process X events forever. */
	XtAppMainLoop(appcontext);

	return 0;
}


/* Definitions for buttons along the bottom left. */

struct button {
	char *label;
	void (*callback)(int);
} button[] = {
	{ "Off", (void (*)())exit },
	{ "Save", save },
	{ NULL, NULL },
};

/* Callback function for buttons. */
static void
button_callback(Widget w, XtPointer client_data, XtPointer call_data)
{
	struct button *b = (struct button *)client_data;

	(*b->callback)(0);
}

/* Card-image data structures. */

#define N_COLS	80
#define N_ROWS	14
#define N_OV	10

typedef struct card {
	unsigned short holes[N_COLS];
	int n_ov[N_COLS];
	char coltxt[N_COLS][N_OV];
	int seq;
	struct card *prev;
	struct card *next;
} card;
static card *ccard;

static int col = 0;
static GC gc, invgc, bggc, corner_gc;
#define CHAR_WIDTH	(fgfontinfo->max_bounds.rbearing + \
			 fgfontinfo->min_bounds.lbearing + 1)
#define CHAR_HEIGHT	(fgfontinfo->ascent + fgfontinfo->descent + 6)
#define CARD_PAD	15
#define CARD_WIDTH	(2*CARD_PAD + (CHAR_WIDTH * N_COLS))
#define CARD_HEIGHT	(2*CARD_PAD + (CHAR_HEIGHT * N_ROWS))
#define CARD_AIR	5

#define	BUTTON_GAP	5
#define BUTTON_BW	2
#define BUTTON_WIDTH	50
#define	BUTTON_HEIGHT	20

static Widget container, porth, scrollw, cardw, posw;

/* Set up the keypunch. */
static void
define_widgets(void)
{
	Dimension w, h;
	Widget ww;
	XtTranslations t;
	XGCValues xgcv;
	int i;
	static char translations[] = "\
		<Key>Left:	Left()\n\
		<Key>BackSpace:	Left()\n\
		<Key>Right:	Right()\n\
		<Key>space:	Right()\n\
		<Key>Home:	Home()\n\
		<Key>Return:	Next()\n\
		<Key>KP_Enter:	Home()\n\
		<Key>Down:	Next()\n\
		<Key>Tab:	Tab()\n\
		<Btn2Down>:	insert-selection(PRIMARY)\n\
		:<Key>:		Data()\n";

	w = CARD_WIDTH + 2*CARD_AIR;
	h = CARD_HEIGHT + 2*CARD_AIR + 2*BUTTON_GAP + 2*BUTTON_BW + BUTTON_HEIGHT;

	/* Create a container for the whole thing. */
	container = XtVaCreateManagedWidget(
	    "container", compositeWidgetClass, toplevel,
	    XtNwidth, w,
	    XtNheight, h,
	    XtNbackground, appres.cabinet,
	    /* XtNforeground, appres.background, */
	    NULL);

	/* Create the porthole within the container. */
	porth = XtVaCreateManagedWidget(
	    "porthole", portholeWidgetClass, container,
	    XtNwidth, CARD_WIDTH,
	    XtNheight, CARD_HEIGHT,
	    XtNx, CARD_AIR,
	    XtNy, CARD_AIR,
	    XtNborderWidth, 0,
	    NULL);

	/* Create scrollable region within the porthole. */
	scrollw = XtVaCreateManagedWidget(
	    "scroll", compositeWidgetClass, porth,
	    XtNwidth, CARD_WIDTH * 3,
	    XtNheight, CARD_HEIGHT * 3,
	    XtNx, -(CARD_PAD+40*CHAR_WIDTH),
	    XtNy, -(2*CARD_HEIGHT) + CARD_PAD,
	    XtNbackground, appres.cabinet,
	    /* XtNforeground, appres.background, */
	    XtNborderWidth, 0,
	    NULL);

	/* Create the card itself. */
	cardw = XtVaCreateManagedWidget(
	    "card", compositeWidgetClass, scrollw,
	    XtNwidth, CARD_WIDTH,
	    XtNheight, CARD_HEIGHT,
	    XtNx, CARD_WIDTH,
	    XtNy, CARD_HEIGHT,
	    XtNborderWidth, 0,
	    XtNbackground, appres.cardcolor,
	    NULL);

	/* Create the buttons in the lower left. */
	for (i = 0; button[i].label; i++) {
		ww = XtVaCreateManagedWidget(
		    button[i].label, commandWidgetClass, container,
		    XtNwidth, BUTTON_WIDTH,
		    XtNheight, BUTTON_HEIGHT,
		    XtNx, BUTTON_GAP + i*(2*BUTTON_BW + BUTTON_WIDTH + BUTTON_GAP),
		    XtNy, CARD_HEIGHT + 2*CARD_AIR + BUTTON_GAP,
		    XtNborderWidth, BUTTON_BW,
		    XtNborderColor, appres.background,
		    NULL
		);
		XtAddCallback(ww, XtNcallback, button_callback,
		    (XtPointer)&button[i]);
	}

	/* Add the position counter in the lower right. */
	posw = XtVaCreateManagedWidget(
	    "pos", labelWidgetClass, container,
	    XtNlabel, "0",
	    XtNwidth, BUTTON_WIDTH,
	    XtNx, w - BUTTON_GAP - BUTTON_WIDTH - 2*BUTTON_BW,
	    XtNy, CARD_HEIGHT + 2*CARD_AIR + BUTTON_GAP,
	    XtNheight, BUTTON_HEIGHT,
	    XtNborderWidth, BUTTON_BW,
	    XtNborderColor, appres.background,
	    XtNresize, False,
	    NULL);

	/* Add the discard button. */
	ww = XtVaCreateManagedWidget(
	    "discard", commandWidgetClass, container,
	    XtNlabel, "Discard",
	    XtNwidth, BUTTON_WIDTH,
	    XtNx, w - 2* (BUTTON_GAP + BUTTON_WIDTH + 2*BUTTON_BW),
	    XtNy, CARD_HEIGHT + 2*CARD_AIR + BUTTON_GAP,
	    XtNheight, BUTTON_HEIGHT,
	    XtNborderWidth, BUTTON_BW,
	    XtNborderColor, appres.background,
	    NULL);
	XtAddCallback(ww, XtNcallback, discard, NULL);

	/* Create graphics contexts for drawing. */
	xgcv.foreground = appres.foreground;
	xgcv.background = appres.cardcolor;
	xgcv.font = fgfontinfo->fid;
	gc = XtGetGC(toplevel, GCForeground|GCBackground|GCFont, &xgcv);
	xgcv.foreground = appres.cabinet;
	corner_gc = XtGetGC(toplevel, GCForeground|GCBackground, &xgcv);
	xgcv.foreground = appres.cardcolor;
	xgcv.background = appres.foreground;
	xgcv.font = fgfontinfo->fid;
	invgc = XtGetGC(toplevel, GCForeground|GCBackground|GCFont, &xgcv);
	xgcv.foreground = appres.foreground;
	xgcv.background = appres.cardcolor;
	xgcv.font = bgfontinfo->fid;
	bggc = XtGetGC(toplevel, GCForeground|GCBackground|GCFont, &xgcv);

	/* Fix the size of the toplevel window. */
	XtVaSetValues(toplevel,
	    XtNwidth, w,
	    XtNheight, h,
	    XtNbaseWidth, w,
	    XtNbaseHeight, h,
	    XtNminWidth, w,
	    XtNminHeight, h,
	    XtNmaxWidth, w,
	    XtNmaxHeight, h,
	    NULL);

	/* Define event translations. */
	t = XtParseTranslationTable(translations);
	XtOverrideTranslations(container, t);
	t = XtParseTranslationTable("<Expose>: Redraw()");
	XtOverrideTranslations(cardw, t);

	/* Inflate it all. */
	XtRealizeWidget(toplevel);
	card_window = XtWindow(cardw);

	/* Add the cursors */
	ww = XtVaCreateManagedWidget(
	    "cursor", compositeWidgetClass, container,
	    XtNwidth, CHAR_WIDTH,
	    XtNheight, CARD_AIR + 20,
	    XtNx, CARD_AIR + CARD_PAD + ((N_COLS/2)+1)*CHAR_WIDTH + 1,
	    XtNy, 0,
	    NULL);
	XtRealizeWidget(ww);
#define C2H (BUTTON_HEIGHT + 2*BUTTON_GAP + 2*BUTTON_BW + 10)
	ww = XtVaCreateManagedWidget(
	    "cursor2", compositeWidgetClass, container,
	    XtNwidth, CHAR_WIDTH,
	    XtNheight, C2H,
	    XtNx, CARD_AIR + CARD_PAD + ((N_COLS/2)+1)*CHAR_WIDTH + 1,
	    XtNy, h-C2H-2,
	    NULL);
	XtRealizeWidget(ww);
}

#include "ascii_hollerith.h"

/* Punch a character into a particular column of the current card. */
static Boolean
punch_char(int cn, unsigned char c)
{
	int j;

	if (c == '\n' || c == '\b' || c > 127 || hollerith[c] == NS)
		return False;

	/* Space?  Do nothing. */
	if (!hollerith[c])
		return True;

	ccard->holes[cn] |= hollerith[c];

	/* Redundant? */
	for (j = 0; j < ccard->n_ov[cn]; j++)
		if (ccard->coltxt[cn][j] == c)
			return True;

	if (ccard->n_ov[cn] < N_OV) {
		ccard->coltxt[cn][ccard->n_ov[cn]] = c;
		++ccard->n_ov[cn];
	}

	return True;
}

/* Render the image of a card column onto the X display. */
/* n_rows is:
*   1 for text
*   2 for text and space
*   3 for text, space, 1 hole
*   4 for text, space, 2 holes
*   ...
*   N_ROWS for all
*/
static void
draw_col(int cn, int row0, int n_rows)
{
	int i;
	int j;
	int x = CARD_PAD + (CHAR_WIDTH * cn);
	int y = CARD_PAD + CHAR_HEIGHT;
	static char no_hole[] = "  0123456789";

	if (!drawing)
		return;

	/* Draw the text at the top, possibly overstruck. */
	if (!row0) {
		Boolean any = False;

		for (j = 0; j < ccard->n_ov[cn]; j++) {
			if (ccard->coltxt[cn][j] < ' ')
				continue;
			XDrawString(display, card_window, gc,
			    x, y, &ccard->coltxt[cn][j], 1);
			any = True;
		}
		if (!any)
			XDrawImageString(display, card_window, gc, x, y, " ", 1);
	}

	/* Draw the holes, or the background numerals. */
	for (j = row0; j < row0 + n_rows; j++) {
		if ((i = j - 2) < 0)
			continue;
		y = CARD_PAD + (CHAR_HEIGHT*(j+1));
		if (ccard->holes[cn] & (1<<i))
			XFillRectangle(display, card_window, bggc,
			    x, y - fgfontinfo->ascent + 1,
			    CHAR_WIDTH-3, CHAR_HEIGHT-7);	/* XXX */
		else
			XDrawImageString(display, card_window, bggc,
			    x, y,
			    &no_hole[i], 1);
	}
}

/* Update the column indicator in the lower right. */
static void
set_posw(int c)
{
	static char bb[3];

	col = c;
	(void) sprintf(bb, "%02d", col+1);
	XtVaSetValues(posw, XtNlabel, bb, NULL);
}

/* Go to the next card. */
static void
do_newcard(int replace)
{
	int i;

	if (!ccard || !replace) {
		card *c;

		c = (card *)XtMalloc(sizeof(card));
		c->prev = ccard;
		c->next = NULL;
		if (ccard)
			ccard->next = c;
		ccard = c;
		c->seq = line_number;
		line_number += 10;
	}
	(void) memset(ccard->coltxt, ' ', sizeof(ccard->coltxt));
	(void) memset(ccard->holes, 0, sizeof(ccard->holes));
	(void) memset(ccard->n_ov, 0, sizeof(ccard->n_ov));
	set_posw(0);
	if (appres.autonumber) {
		char ln_buf[9];

		(void) sprintf(ln_buf, "%08d", ccard->seq);
		for (i = 0; i < 8; i++)
			punch_char(72+i, ln_buf[i]);
	}
}

/* Redraw the entire card image. */
static void
redraw(Widget wid, XEvent  *event, String  *params, Cardinal *num_params)
{
	int i;
	Dimension x, y, w, h;
	int row0, nrows;

	if (event && event->type == Expose) {
		x = event->xexpose.x;
		y = event->xexpose.y;
		w = event->xexpose.width;
		h = event->xexpose.height;
	} else {
		x = y = 0;
		w = CARD_WIDTH;
		h = CARD_HEIGHT;
	}

	/* Redisplay the clipped corner. */
	if (x <= 15 || y <= 20) {
		static XPoint p[4] = {
			{ 0, 0 },
			{ 15, 0 },
			{ 0, 20 },
			{ 0, 0 }
		};

		XFillPolygon(display, card_window, corner_gc, p, 4,
		    Convex, CoordModeOrigin);
	}

	/* Slice off the padding. */
	if (x < CARD_PAD) {			/* Left. */
		if (w <= CARD_PAD - x)
			return;
		w -= CARD_PAD - x;
		x = 0;
	} else
		x -= CARD_PAD;
	if (y < CARD_PAD) {			/* Top. */
		if (h <= CARD_PAD - y)
			return;
		h -= CARD_PAD - y;
		y = 0;
	} else
		y -= CARD_PAD;
	if (x >= (CHAR_WIDTH * N_COLS))		/* Right. */
		return;
	if (x + w > (CHAR_WIDTH * N_COLS))
		w = (CHAR_WIDTH * N_COLS) - x;
	if (y >= (CHAR_HEIGHT * N_ROWS))	/* Bottom. */
		return;
	if (y + h > (CHAR_HEIGHT * N_ROWS))
		h = (CHAR_HEIGHT * N_ROWS) - y;
	
	row0 = y / CHAR_HEIGHT;
	nrows = ((y+h+(CHAR_HEIGHT-1)) / CHAR_HEIGHT) - row0;
	if (!nrows)
		nrows++;
	for (i = (x / CHAR_WIDTH);
	     i < (x + w + (CHAR_WIDTH-1)) / CHAR_WIDTH;
	     i++)
		draw_col(i, row0, nrows);
}

/* Exit. */
static void
delete_window(Widget wid, XEvent *event, String *params, Cardinal *num_params)
{
	exit(0);
}

/* Find the first card in the deck. */
static card *
first_card(void)
{
	card *c;

	for (c = ccard; c && c->prev; c = c->prev)
		;
	return c;
}

static void
save(int ignored)
{
	save_popup();
}

/*
 * Internals of functions that are enqueued with a delay.
 */

static void
do_nothing(int ignored)
{
}

static void
do_data(int c)
{
	extern void do_right(int);

	if (col < N_COLS && punch_char(col, c)) {
		draw_col(col, 0, N_ROWS);
#if defined(SOUND) /*[*/
		loud_click();
#endif /*]*/
		do_right(0);
	}
}

static void
do_multipunch(int c)
{
	extern void do_right(int);

	if (col < N_COLS && punch_char(col, c)) {
		draw_col(col, 0, N_ROWS);
#if defined(SOUND) /*[*/
		loud_click();
#endif /*]*/
	}
}

static void
do_left(int c)
{
	extern void do_pan_left(int);

	if (col) {
		do_pan_left(0);
		set_posw(col - 1);
	}
}

static void
do_right(int do_click)
{
	extern void do_pan_right(int);

	if (col < N_COLS-1) {
		do_pan_right(do_click);
		set_posw(col + 1);
	}
}

static void
do_pan_left(int ignored)
{
	Dimension x;

	XtVaGetValues(scrollw, XtNx, &x, NULL);
	x += CHAR_WIDTH;
	XtVaSetValues(scrollw, XtNx, x, NULL);
#if defined(SOUND) /*[*/
	soft_click();
#endif /*]*/
}

static void
do_pan_right(int do_click)
{
	Dimension x;

	XtVaGetValues(scrollw, XtNx, &x, NULL);
	x -= CHAR_WIDTH;
	XtVaSetValues(scrollw, XtNx, x, NULL);
#if defined(SOUND) /*[*/
	if (do_click)
		soft_click();
#endif /*]*/
}

static void
do_pan_up(int ignored)
{
	Dimension y;

	XtVaGetValues(scrollw, XtNy, &y, NULL);
	y += CHAR_HEIGHT;
	XtVaSetValues(scrollw, XtNy, y, NULL);
}

static void
do_home(int ignored)
{
	do_pan_left(0);
	set_posw(col - 1);
}

static void
do_slam(int ignored)
{
	XtVaSetValues(scrollw,
	    XtNx, -(CARD_PAD+40*CHAR_WIDTH),
	    XtNy, -(2*CARD_HEIGHT) + CARD_PAD,
	    NULL);
}

/*
 * Event queueing: Inserting artificial delays in processing certain events.
 */
#define N_EVENTS	500
enum evtype { DUMMY, DATA, MULTI, LEFT, RIGHT, HOME,
	      PAN_RIGHT, PAN_LEFT, PAN_UP, SLAM, NEWCARD };
void (*eq_fn[])(int) = {
	do_nothing,
	do_data,
	do_multipunch,
	do_left,
	do_right,
	do_home,
	do_pan_right,
	do_pan_left,
	do_pan_up,
	do_slam,
	do_newcard
};
struct {
	enum evtype evtype;
	unsigned char param;	/* optional */
	int delay;
} eventq[N_EVENTS];
int eq_fillp, eq_emptyp, eq_count;

static void
deq_event(XtPointer data, XtIntervalId *id)
{
	int delay;

	if (!eq_count)
		return;
	(*eq_fn[eventq[eq_emptyp].evtype])(eventq[eq_emptyp].param);
	delay = eventq[eq_emptyp].delay;
	eq_emptyp = (eq_emptyp + 1) % N_EVENTS;
	if (--eq_count)
		(void) XtAppAddTimeOut(appcontext, delay, deq_event, NULL);
}

static Boolean
enq_event(enum evtype evtype, unsigned char param, Boolean restricted, int delay)
{
	if (restricted && eq_count) {
		XBell(display, 0);
		return False;
	}
	if (eq_count >= N_EVENTS) {
		XBell(display, 0);
		return False;
	}
	if (evtype == DUMMY)
		return True;
	eventq[eq_fillp].evtype = evtype;
	eventq[eq_fillp].param = param;
	eventq[eq_fillp].delay = delay;
	eq_fillp = (eq_fillp + 1) % N_EVENTS;
	if (!eq_count++)
		(void) XtAppAddTimeOut(appcontext, delay, deq_event, NULL);
	return True;
}

/*
 * Externals of delayed functions, called by the toolkit.
 */

static void
data(Widget wid, XEvent *event, String *params, Cardinal *num_params)
{
	XKeyEvent *kevent = (XKeyEvent *)event;
	char buf[10];
	KeySym ks;
	int ll;

	ll = XLookupString(kevent, buf, 10, &ks, (XComposeStatus *)NULL);
	if (ll == 1) {
		if (kevent->state & (Mod1Mask | Mod2Mask))
			enq_event(MULTI, buf[0], !appres.typeahead, SLOW);
		else
			enq_event(DATA, buf[0], !appres.typeahead, SLOW);
	}
}

static void
left(Widget wid, XEvent *event, String *params, Cardinal *num_params)
{
	enq_event(LEFT, 0, !appres.typeahead, SLOW);
}

static void
right(Widget wid, XEvent *event, String *params, Cardinal *num_params)
{
	enq_event(RIGHT, 1, !appres.typeahead, SLOW);
}

static void
home(Widget wid, XEvent *event, String *params, Cardinal *num_params)
{
	int i;

	if (!enq_event(DUMMY, 0, True, SLOW))
		return;

	for (i = 0; i < col; i++)
		enq_event(HOME, 0, False, FAST);
}

static void
next(Widget wid, XEvent *event, String *params, Cardinal *num_params)
{
	if (ccard->next) {
		ccard = ccard->next;
		set_posw(0);
	} else
		new_card();
}

static void
tab(Widget wid, XEvent *event, String *params, Cardinal *num_params)
{
	int i;

	if (!enq_event(DUMMY, 0, True, SLOW))
		return;

	for (i = col; i < 6; i++)
		enq_event(RIGHT, 1, False, SLOW);
}

/* Throw away this card. */
static void
discard(Widget w, XtPointer client_data, XtPointer call_data)
{
	int i;

	if (!enq_event(DUMMY, 0, True, SLOW))
		return;

	/* Scroll the previous card away. */
	if (ccard) {
		for (i = 0; i <= col; i++)
			enq_event(LEFT, 0, False, FAST);
		for (i = 0; i < N_COLS/2; i++)
			enq_event(PAN_LEFT, 0, False, FAST);
	}

	enq_event(NEWCARD, True, False, FAST);

	/* Scroll the new card down. */
	enq_event(SLAM, 0, False, SLOW);
	for (i = 0; i <= N_ROWS; i++)
		enq_event(PAN_UP, 0, False, SLOW);

}

/* Generate a fresh card. */
static void
new_card(void)
{
	int i;

	/* Scroll the previous card away. */
	if (ccard) {
		for (i = col; i < N_COLS; i++)
			enq_event(RIGHT, 0, False, FAST);
		for (i = 0; i < N_COLS/2 + 2; i++)
			enq_event(PAN_RIGHT, 0, False, FAST);
	}

	enq_event(NEWCARD, False, False, FAST);

	/* Scroll the new card down. */
	enq_event(SLAM, 0, False, SLOW);
	for (i = 0; i <= N_ROWS; i++)
		enq_event(PAN_UP, 0, False, SLOW);
}

#define NP	5
Atom paste_atom[NP];
int n_pasting = 0;
int pix = 0;
Time paste_time;

/* Pasting support. */
static void
paste_callback(Widget w, XtPointer client_data, Atom *selection, Atom *type,
    XtPointer value, unsigned long *length, int *format)
{
	char *s;
	unsigned long len;

	if ((value == NULL) || (*length == 0)) {
		XtFree(value);

		/* Try the next one. */
		if (n_pasting > pix)
			XtGetSelectionValue(w, paste_atom[pix++], XA_STRING,
			    paste_callback, NULL, paste_time);
		return;
	}

	s = (char *)value;
	len = *length;
	while (len--) {
		if (*s == '\n')
			break;
		if (*s < ' ' || (*s & 0x80)) {
			s++;
			continue;
		}
		if (!enq_event(DATA, *s++, False, SLOW))
			break;
	}
	n_pasting = 0;

	XtFree(value);
}

static void
insert_selection(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
	int	i;
	Atom	a;
	XButtonEvent *bevent = (XButtonEvent *)event;

	n_pasting = 0;
	for (i = 0; i < *num_params; i++) {
		a = XInternAtom(display, params[i], True);
		if (a == None) {
			XtWarning("no atom for selection");
			continue;
		}
		if (n_pasting < NP)
			paste_atom[n_pasting++] = a;
	}
	pix = 0;
	if (n_pasting > pix) {
		paste_time = bevent->time;
		XtGetSelectionValue(w, paste_atom[pix++], XA_STRING,
		    paste_callback, NULL, paste_time);
	}
}

/* 'Save' pop-up. */

static Boolean save_popup_created = False;
static Widget save_shell, save_dialog;

static void
center_it(Widget w, XtPointer client_data, XtPointer call_data)
{
	Dimension width, height;
	Position toplevel_x, toplevel_y;
	Dimension toplevel_width, toplevel_height;

	/* Get the popup's dimensions */
	XtVaGetValues(w,
	    XtNwidth, &width,
	    XtNheight, &height,
	NULL);

	/* Get the toplevel position and dimensions. */
	XtVaGetValues(toplevel,
	    XtNwidth, &toplevel_width,
	    XtNheight, &toplevel_height,
	    XtNx, &toplevel_x,
	    XtNy, &toplevel_y,
	    NULL);

	/* Compute our position relatve to those. */
	XtVaSetValues(w,
	    XtNx, toplevel_x + (toplevel_width - width) / 2,
	    XtNy, toplevel_y + (toplevel_height - height) / 2,
	    NULL);
}

static void
save_file(void)
{
	char *sfn;
	FILE *f;
	card *c;
	int i, j, h;

	XtVaGetValues(save_dialog, XtNvalue, &sfn, NULL);
	f = fopen(sfn, "w");
	if (f == NULL) {
	    XBell(display, 100);
	    XtVaSetValues(save_dialog,
		XtNlabel, strerror(errno),
		NULL);
	    XtVaSetValues(XtNameToWidget(save_dialog, XtNlabel),
		XtNforeground, appres.errcolor,
		NULL);
	    return;
	}
	XtPopdown(save_shell);

	for (c = first_card(); c; c = c->next) {
		for (i = 0; i < N_COLS; i++) {
			if (!c->n_ov[i]) {
				fputc(' ', f);
				continue;
			}
			/* Try to find a single character. */
			for (h = 0; h < 128; h++) {
				if (h == '\n' || h == '\b')
					continue;
				if (hollerith[h] != NS &&
				    hollerith[h] == c->holes[i]) {
					fputc(h, f);
					break;
				}
			}
			if (h < 128)
				continue;
			/* Write it with backspaces. */
			for (j = 0; j < c->n_ov[i]; j++) {
				if (j)
					fputc('\b', f);
				fputc(c->coltxt[i][j], f);
			}
		}
		fputc('\n', f);
	}
	fclose(f);
}

static void
save_callback(Widget w, XtPointer client_data, XtPointer call_data)
{
	save_file();
}

static void
cancel_callback(Widget w, XtPointer client_data, XtPointer call_data)
{
	XtPopdown(save_shell);
}

static void
confirm(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
	save_file();
}

static void
save_popup(void)
{
	Widget w;

	if (!save_popup_created) {
		Dimension width, height;

		/* Create the shell. */
		save_shell = XtVaCreatePopupShell("save",
		    transientShellWidgetClass, toplevel,
		    NULL);
		XtAddCallback(save_shell, XtNpopupCallback, center_it, NULL);

		/* Create the dialog in the popup. */
		save_dialog = XtVaCreateManagedWidget(
		    "dialog", dialogWidgetClass,
		    save_shell,
		    XtNvalue, "",
		    XtNbackground, appres.cabinet,
		    NULL);
		w = XtNameToWidget(save_dialog, XtNvalue);
		XtVaSetValues(w,
		    XtNwidth, 200,
		    NULL);
		XtOverrideTranslations(w,
		    XtParseTranslationTable("<Key>Return: confirm()"));
		XtVaSetValues(XtNameToWidget(save_dialog, XtNlabel),
		    XtNwidth, 200,
		    XtNbackground, appres.cabinet,
		    NULL);
		if (appres.cabinet == XBlackPixel(display, default_screen))
			XtVaSetValues(XtNameToWidget(save_dialog, XtNlabel),
			    XtNforeground, XWhitePixel(display, default_screen),
			    NULL);

		/* Add 'Save' and 'Cancel' buttons to the dialog. */
		w = XtVaCreateManagedWidget("Save", commandWidgetClass,
		    save_dialog, NULL);
		XtAddCallback(w, XtNcallback, save_callback, NULL);
		w = XtVaCreateManagedWidget("Cancel", commandWidgetClass,
		    save_dialog, NULL);
		XtAddCallback(w, XtNcallback, cancel_callback, NULL);

		/* Set the width and height. */
		XtRealizeWidget(save_shell);
		XtVaGetValues(save_shell,
		    XtNwidth, &width,
		    XtNheight, &height,
		NULL);
		XtVaSetValues(save_shell,
		    XtNheight, height,
		    XtNwidth, width,
		    XtNbaseHeight, height,
		    XtNbaseWidth, width,
		    XtNminHeight, height,
		    XtNminWidth, width,
		    XtNmaxHeight, height,
		    XtNmaxWidth, width,
		    NULL);

		save_popup_created = True;
	}
	XtVaSetValues(save_dialog,
	    XtNlabel, "Save File Name",
	    NULL);
	if (appres.cabinet == XBlackPixel(display, default_screen))
		XtVaSetValues(XtNameToWidget(save_dialog, XtNlabel),
		    XtNforeground, XWhitePixel(display, default_screen),
		    NULL);
	else
		XtVaSetValues(XtNameToWidget(save_dialog, XtNlabel),
		    XtNforeground, appres.foreground,
		    NULL);
	XtPopup(save_shell, XtGrabExclusive);
	/*XSetWMProtocols(display, XtWindow(save_shell), &delete_me, 1);*/
}

